@@ -216,18 +216,25 @@ def coords2pdb(seq, coords, cloud_mask, prefix="", name="af2_struct.pdb"):
 
 # sidechainnet / other data utils
 
-def get_esm_embedd(seq, embedd_model, batch_converter):
+def get_esm_embedd(seq, embedd_model, batch_converter, embedd_type="per_tok"):
     """ Returns the ESM embeddings for a protein. 
         Inputs: 
         * seq: (L,) tensor of ints (in sidechainnet int-char convention)
-        * embedd_model: ESM model
-        * batch_converter: ESM batch converter
+        * embedd_model: ESM model (see train_end2end.py for an example)
+        * batch_converter: ESM batch converter (see train_end2end.py for an example)
+        * embedd_type: one of ["mean", "per_tok"]. 
+                       "per_tok" is recommended if working with sequences.
     """
     str_seq = "".join([VOCAB._int2char[x]for x in seq.cpu().numpy()])
     batch_labels, batch_strs, batch_tokens = batch_converter( [(0, str_seq)] )
     with torch.no_grad():
         results = embedd_model(batch_tokens, repr_layers=[33], return_contacts=False)
-    return results["representations"][33].to(seq.device)
+    # index 0 is for start token. so take from 1 one
+    token_reps = results["representations"][33][ 1 : len(str_seq) + 1].to(seq.device)
+    if embedd_type == "mean":
+        token_reps = token_reps.mean(dim=0)
+    return token_reps
+
 
 
 def get_all_protein_ids(dataloader, verbose=False):
@@ -680,24 +687,27 @@ def calc_phis_numpy(pred_coords, N_mask, CA_mask, C_mask=None,
 # alignment by centering + rotation to compute optimal RMSD
 # adapted from : https://github.com/charnley/rmsd/
 
-def kabsch_torch(X, Y):
+def kabsch_torch(X, Y, cpu=True):
     """ Kabsch alignment of X into Y. 
         Assumes X,Y are both (Dims x N_points). See below for wrapper.
     """
+    device = X.device
     #  center X and Y to the origin
     X_ = X - X.mean(dim=-1, keepdim=True)
     Y_ = Y - Y.mean(dim=-1, keepdim=True)
     # calculate convariance matrix (for each prot in the batch)
-    C = torch.matmul(X_, Y_.t())
+    C = torch.matmul(X_, Y_.t()).detach()
+    if cpu: 
+        C = C.cpu()
     # Optimal rotation matrix via SVD - warning! W must be transposed
-    V, S, W = torch.svd(C.detach())
+    V, S, W = torch.svd(C)
     # determinant sign for direction correction
     d = (torch.det(V) * torch.det(W)) < 0.0
     if d:
         S[-1]    = S[-1] * (-1)
         V[:, -1] = V[:, -1] * (-1)
     # Create Rotation matrix U
-    U = torch.matmul(V, W.t())
+    U = torch.matmul(V, W.t()).to(device)
     # calculate rotations
     X_ = torch.matmul(X_.t(), U).t()
     # return centered and aligned
@@ -729,6 +739,23 @@ def kabsch_numpy(X, Y):
 
 # metrics - more formulas here: http://predictioncenter.org/casp12/doc/help.html
 
+def distmat_loss_torch(X, Y, p=2, q=2, distmat_mask=None):
+    """ Calculates a loss on the distance matrix - no need to align structs.
+        Inputs: 
+        * X: (N, d) tensor. the predicted stricture 
+        * Y: (N, d) tensor. the true structure
+        * p: int. power for the distance calculation (2 for euclidean)
+        * q: float. power for the scaling of the loss (2 for MSE, 1 for MAE, etc)
+        * distmat_mask: (N, N) mask (boolean or weights for each ij pos). optional.
+    """
+    # **2 ensures always positive. Later scale back to desired power
+    loss = ( torch.cdist(X, X, p=p) - torch.cdist(Y, Y, p=p) )**2 
+    if q != 2:
+        loss = loss**(q/2)
+    if distmat_mask is not None:
+        loss *= distmat_mask.float()
+    return loss.mean()
+
 def rmsd_torch(X, Y):
     """ Assumes x,y are both (B x D x N). See below for wrapper. """
     return torch.sqrt( torch.mean((X - Y)**2, axis=(-1, -2)) )
